package nl.uvt.locator;

import java.net.URL;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import nl.uvt.commons.io.*;

public class MerescoConnection {

	class MerescoCQL extends CQLfields{

		public void fill() {
			currentPrefix = "meresco.";
			String val = Locator.locatorConfig.getProperty( currentPrefix+"issnSelector", "mods.identifier=" );
			issnSelector = val.replaceAll("\"","");
			val = Locator.locatorConfig.getProperty( currentPrefix+"isbnSelector", "mods.identifier=" );
			isbnSelector = val.replaceAll("\"","");
			val = Locator.locatorConfig.getProperty( currentPrefix+"epnSelector", "mods.location.holdingSimple.copyInformation.note=" );
			epnSelector = val.replaceAll("\"","");
			val = Locator.locatorConfig.getProperty( currentPrefix+"titleSelector", "mods.titleInfo.title=" );
			titleSelector = val.replaceAll("\"","");
			val = Locator.locatorConfig.getProperty( currentPrefix+"subTitleSelector", "mods.titleInfo.subTitle=" );
			subTitleSelector = val.replaceAll("\"","");
			val = Locator.locatorConfig.getProperty( currentPrefix+"fullTitleSelector",  "norm.fulljournaltitle exact" );
			fullTitleSelector = val.replaceAll("\"","");
			val = Locator.locatorConfig.getProperty( currentPrefix+"locationSelector", "mods.location.holdingSimple.copyInformation.shelfLocator=" );
			locationSelector = val.replaceAll("\"","");
			val = Locator.locatorConfig.getProperty( currentPrefix+"authorSelector", "norm.name.lastname=" );
			authorSelector = val.replaceAll("\"","");
			val = Locator.locatorConfig.getProperty( currentPrefix+"repositorySelector", "meta.repository.repositoryGroupId exact opc-uvt-nl" );
			repositorySelector = val.replaceAll("\"","");
		}

	}
	
	MerescoCQL cqlData = new MerescoCQL();
	

	List<LocatorInfo> query( QueryData pQ ) throws LocatorException {
		cqlData.fill();
		List<LocatorInfo> meResult = null;
		nu.xom.Document result;
		String baseUrl = Locator.locatorConfig.getProperty("merescoUri");
		try {
			URL url = new URL( baseUrl );
			Properties props = new Properties( Locator.sruDefaultProps );
			props.setProperty("recordSchema","mods" );
			String query = makeCQL( pQ, cqlData );
			if ( query.length() == 0 ){
				// fails when no useful data available
				return meResult;
			}
			query += " and " + cqlData.repositorySelector; 
			props.setProperty("query", query );
			result = Http.httpGetDocument( url, props, Locator.httpConfig );
			if ( result != null ){
				int hits = Locator.extractHits( result );
				Locator.logger.debug("meresco numOfRecords=" + hits );
				if ( hits > 0 ){
					//					Locator.logger.debug( "meresco doc:" + result.toXML());
					String sheetName = Locator.locatorConfig.getProperty("merescoSheet");
					meResult = HoldingInfo.getModsResults( result, pQ, sheetName );
				}
			}
			else {
				Locator.logger.warn("No response from:" + url );
			}
		}
		catch ( Exception e) {
			throw new LocatorException( "meresco URL: '" + baseUrl + "' failed"); 
		}
		return meResult;
	}

	public static String makeCQL( QueryData pQ, CQLfields cqlData ){
		String query = "";
		if ( !pQ.epn.equals("") ) {
			Locator.logger.debug("direct search with EPN: " + pQ.epn );
			query = cqlData.epnSelector + pQ.epn;
		}
		else if ( pQ.genre.equals( "article") ){
			if ( pQ.issn.length() > 0 ){
				query = cqlData.issnSelector + pQ.issn;
				if ( pQ.location.length() > 0 ){
					String location = fixLocation( pQ.location );
					query += " and " + cqlData.locationSelector + "\"" + location + "\"";
				}
			}
			else if ( pQ.title.length() > 0 ){
				StringBuffer main = new StringBuffer();
				StringBuffer sub = new StringBuffer();
				splitTitle( pQ.title, main, sub );
				String title = main.toString();
				// we don't want spurious " in our query (tucat seem to swallow it, but it is invalid CQL)
				title = filterTitle( title );
				// index is case insensitive and stopwords are assumed lowercase. 
				title = title.toLowerCase();
				title = filterStopWords( title );
				if ( cqlData.currentPrefix.equals("tucat.")	)
					title = picaFixTitle( title );
				query = cqlData.fullTitleSelector + "\"" + title + "\"";
			}
		}
		else if ( pQ.isbn.length() > 0 ){
			Locator.logger.debug("ISBN: " + pQ.isbn );
			if ( pQ.isbn.length() == 13 ){
				Locator.logger.debug("ISBN: is lang genoeg " + pQ.isbn.length() );
				query = cqlData.isbnSelector + pQ.isbn;
			}
			else if ( pQ.isbn.length() == 10 ){
				Locator.logger.debug("ISBN verlengen " );
				String isbn13 = to13( pQ.isbn );
				query = cqlData.isbnSelector + pQ.isbn + " or " + cqlData.isbnSelector + isbn13;						
			}
		}
		else if ( pQ.location.length() > 0 ){
			String location = fixLocation( pQ.location );
			query = cqlData.locationSelector + "\"" + location + "\"";
		}
		else {
			String fullTitle = null;
			String mainTitle = "";
			String subTitle = "";
			if ( pQ.title.length() > 0 ) {
				fullTitle = pQ.title;
				StringBuffer main = new StringBuffer();
				StringBuffer sub = new StringBuffer();
				splitTitle( pQ.title, main, sub );
				mainTitle = main.toString();
				subTitle = sub.toString();
			}
			else if ( pQ.atitle.length() > 0 ) {
				fullTitle = pQ.atitle;
				mainTitle = pQ.atitle;
			}
			if ( cqlData.currentPrefix.equals("tucat.")){
				fullTitle = picaFixTitle( fullTitle );
				mainTitle = picaFixTitle( mainTitle );
				subTitle = picaFixTitle( subTitle );
			}
			String author = null;
			if ( !pQ.genre.equals("bookitem") && 	
					!pQ.genre.equals("proceeding") &&
					pQ.aulast.length() > 0 ){
				author = cqlData.authorSelector + "\"" + pQ.aulast + "\""; 
			}
			if ( fullTitle != null ) {
				if ( pQ.genre.equals("bookitem")){
					query = cqlData.fullTitleSelector + "\"" + fullTitle + "\"";
					if ( author != null )
						query += " and " + author;
				}
				else if ( !mainTitle.equals("") ){
					if ( subTitle.equals("") ){
						query = cqlData.titleSelector + "\"" + mainTitle + "\"";
					}
					else {
						query = cqlData.titleSelector + "\"" + mainTitle + "\" and " + cqlData.subTitleSelector + "\"" + subTitle + "\"";
					}
					if ( author != null )
						query += " and " + author;
				}
				else if ( author != null )
					query = author;
			}
			else if ( author != null )
				query = author;
		}
		return query;
	}


	public static String picaFixTitle (String title) {
		// pica supports only strings of 64 chars, it seems
		String result = title;
		result = result.replace('"', ' ');
		if ( title.length() > 64 ){
			result = result.replace('(', ' ');
			result = result.replace(')', ' ');
			result = result.replace('[', ' ');
			result = result.replace(']', ' ');
			result = result.replace('{', ' ');
			result = result.replace('}', ' ');
			int pos = result.lastIndexOf(' ');
			while ( pos > 0 && pos > 64 ){
				pos = result.lastIndexOf(' ', pos-1 );
			}
			if ( pos == 0 ){
				pos = result.indexOf(':');
			}
			if ( pos > 0 )
				result = result.substring(0,pos);
			Locator.logger.debug("fixed up title from '" + title + "' to '" + result + "'");
		}
		return result;
	}

	static String to13(String isbn) {
		Locator.logger.debug( "ISBN10to13 :"+ isbn );
		if ( isbn.length() < 10 ){
			Locator.logger.debug( "isbn is to short!");
			return isbn;
		}
		String NUMVALUES = "0123456789";
		String work = isbn.substring(0, 9);
		work = "978" + work;
		char[] bytes = work.toCharArray();
		boolean three = false;
		int checksum = 0;
		for ( int i=0; i < 12; ++i ){
			if ( three )
				checksum += 3 * NUMVALUES.indexOf(bytes[i]);
			else
				checksum += 1 * NUMVALUES.indexOf(bytes[i]);
			three = !three;
		}
		int rem = checksum % 10;
		int last = 0;
		if ( rem != 0 )
			last = 10 - rem;
		String result = new String(bytes) + last;
		Locator.logger.debug( "ISBN10to13 gives "+ result );
		return result;
	}

	private static void splitTitle(String title, StringBuffer main,	StringBuffer sub) {
		String titleSep = Locator.locatorConfig.getProperty( "subTitleSeparator", "DOOD DOOD" );
		titleSep = titleSep.replaceAll("\"","");
		Locator.logger.debug("separator = '"+titleSep+"'" );
		String titleParts[] = title.split( titleSep, 2 );
		if ( titleParts.length > 1 ){
			for( int i=0; i < titleParts.length; ++i ){
				titleParts[i] =	titleParts[i].trim();
				Locator.logger.debug( "splitTitle part["+i+"]= '"+titleParts[i] +"'");
			}
			main.append( titleParts[0] );
			sub.append( titleParts[1] );
		}
		else
			main.append( title );
	}

	static String fixLocation( String loc ) {
		String result = loc.replaceAll("/\\d{4}.*", "/...");
		if ( !loc.equals(result))
			Locator.logger.debug("fixLocation replaced '" + loc + "' by: '" + result + "'" );			
		return result;
	}

	static String filterTitle( String query ) {
		String result = query;
		if ( Locator.filterChars == null ){
			Locator.filterChars = Locator.locatorConfig.getProperty("filterChars","[\"]");
		}
		if ( Locator.filterChars != null ){
			try {
				result = result.replaceAll( Locator.filterChars, "");
			}
			catch ( Exception e ){
				Locator.logger.warn( "filterTitle failed! Something wrong with filterChars regExp? " + Locator.filterChars );  
			}
		}
		if ( !result.equals(query) )
			Locator.logger.debug("removed unwanted chars " + Locator.filterChars + " in '" + query +"' ==> '" + result +"'" );
		return result;
	}

	static String filterStopWords( String query ) {
		String result = query;
		if ( Locator.stopWords == null ){
			Locator.stopWords = new HashSet<String>();
			String words = Locator.locatorConfig.getProperty("stopWords","");
			if ( words.equals("") ){
				// hardwired default
				Locator.logger.warn("PLEASE FIX THIS: NO stopWords found in configfile." );
				Locator.logger.warn("PLEASE FIX THIS: we don't use stopwords." );
			}
			String [] doiS = words.split(",");
			for ( int i=0; i < doiS.length; ++i ){
				Locator.stopWords.add(doiS[i].trim());
			}
			Locator.logger.debug("filled Known stopWords list: " + Locator.stopWords.toString() );
		}
		Iterator<String> it = Locator.stopWords.iterator();
		while ( it.hasNext() ){
			result = result.replaceFirst("^" + it.next() + " ", "");
		}
		if ( !result.equals(query) )
			Locator.logger.debug("replaced stopwords in '" + query +"' ==> '" + result +"'" );
		return result;
	}


}
